import areaofinterest
import convolution
import header
import logger
import numpy as np
import torch
import torch.nn as nn
import utility

class Darknet(nn.Module):
    def __init__(self, file_path_name_config, file_path_name_weights) -> None:
        # Call parent constructor
        super(Darknet, self).__init__()

        # Load configurations, weights, and create modules
        configurations = self.load_configurations(file_path_name_config)
        weights = self.load_weights(file_path_name_weights)
        modules = self.create_modules(configurations, weights)

        # Store configurations and modules to member variables
        self.configurations = configurations
        self.network_information = modules[0]
        self.weights_header = modules[1]
        self.module_list = modules[2]
        self.layer_output_caching_flags = modules[3]

        # Log trace messages
        logger.log_trace(self.configurations)
        logger.log_trace(self.network_information)
        logger.log_trace(self.module_list)
        logger.log_trace(self.layer_output_caching_flags)
        logger.log_trace(sum(self.layer_output_caching_flags))

        return

    def get_configurations(self):
        # Return configurations
        return self.configurations

    @staticmethod
    def load_configurations(file_path_name):
        # Open configuration file
        file = open(file_path_name, "r")

        # Read lines
        lines = file.readlines()
        lines = [line for line in lines if line[0] != '#']  # Remove comments
        lines = [line.strip() for line in lines]  # Remove white spaces
        lines = [line for line in lines if line != ""]    # Remove empty lines

        # Declare configuration dictionary and list
        configuration = {}
        configurations = []

        # Process lines
        for line in lines:
            # If the current line is layer type
            if line[0] == '[':
                # Append last layer configuration if not empty
                if len(configuration) != 0:
                    configurations.append(configuration)
                    configuration = {}

                # Store current layer type
                configuration["type"] = line.strip("[]")
            # If the current line is regular
            else:
                # Store parameter
                parameter = line.split('=')
                parameter = [i.strip() for i in parameter]  # Remove white spaces
                configuration[parameter[0]] = parameter[1]

        # Append last layer if not empty
        if len(configuration) != 0:
            configurations.append(configuration)

        # Return configurations
        return configurations

    @staticmethod
    def load_weights(file_path_name):
        # Open weights file
        file = open(file_path_name, "r")

        # Read weights header and data
        # The first 5 values are header
        # 1: major version number
        # 2: minor version number
        # 3: patch version number
        # 4 & 5: images seen by the network during training
        weights_header = np.fromfile(file, dtype = np.int32, count = 5)
        weights_data = np.fromfile(file, dtype = np.float32)

        # Return weights
        return weights_header, weights_data

    @staticmethod
    def create_modules(configurations, weights):
        # Declare variables
        aoi = areaofinterest.AoI(header.focusification_mask)
        module_list = nn.ModuleList()
        network_information = configurations[0]
        input_channel_size_next = header.input_size_channel # Input channel size of the next layer
        input_channel_sizes = []
        layer_output_caching_flags = [False] * len(configurations[1:])
        weights_header = torch.from_numpy(weights[0])
        weights_data = weights[1]
        weights_index = 0

        # Append next input channel size
        input_channel_sizes.append(input_channel_size_next)

        # Process configurations
        for layer_index, configuration in enumerate(configurations[1:]):
            # Declare sequential module
            module = nn.Sequential()

            # Process convolutional configurations
            if configuration["type"] == "convolutional":
                # Declare batch normalization and pad flag
                batch_normalize = False
                pad = 0

                # Read batch normalization flag
                if "batch_normalize" in configuration:
                    batch_normalize = bool(configuration["batch_normalize"])

                # Read filters, size, and stride
                filters = int(configuration["filters"])
                size = int(configuration["size"])
                stride = int(configuration["stride"])

                # Read pad
                if bool(configuration["pad"]):
                    pad = (size - 1) // 2

                # Read activation
                activation = configuration["activation"]

                # Create convolutional module
                if header.focusify and layer_index > header.layer_index_focusify_start and layer_index < header.layer_index_focusify_end:
                    conv2d = convolution.FocusedConv2d(input_channel_sizes[layer_index], filters, size,
                        stride, pad, bias = not batch_normalize, aoi = aoi)
                else:
                    conv2d = nn.Conv2d(input_channel_sizes[layer_index], filters, size,
                        stride, pad, bias = not batch_normalize)

                # Create batch normalization module
                if batch_normalize:
                    batchnorm2d = nn.BatchNorm2d(filters)

                    # Load bias data for batch normalization module
                    bias_size = batchnorm2d.bias.numel()
                    bias_data = torch.from_numpy(weights_data[weights_index:(weights_index + bias_size)])
                    bias_data = bias_data.view_as(batchnorm2d.bias.data)
                    batchnorm2d.bias.data.copy_(bias_data)
                    weights_index += bias_size

                    # Load weight data for batch normalization module
                    weight_size = batchnorm2d.weight.numel()
                    weight_data = torch.from_numpy(weights_data[weights_index:(weights_index + weight_size)])
                    weight_data = weight_data.view_as(batchnorm2d.weight.data)
                    batchnorm2d.weight.data.copy_(weight_data)
                    weights_index += weight_size

                    # Load running mean data for batch normalization module
                    runnning_mean_size = batchnorm2d.running_mean.numel()
                    runnning_mean_data = torch.from_numpy(weights_data[weights_index:(weights_index + runnning_mean_size)])
                    runnning_mean_data = runnning_mean_data.view_as(batchnorm2d.running_mean.data)
                    batchnorm2d.running_mean.data.copy_(runnning_mean_data)
                    weights_index += runnning_mean_size

                    # Load running variance data for batch normalization module
                    running_var_size = batchnorm2d.running_var.numel()
                    running_var_data = torch.from_numpy(weights_data[weights_index:(weights_index + running_var_size)])
                    running_var_data = running_var_data.view_as(batchnorm2d.running_var.data)
                    batchnorm2d.running_var.data.copy_(running_var_data)
                    weights_index += running_var_size
                # Load bias data for convolutional module
                else:
                    bias_size = conv2d.bias.numel()
                    bias_data = torch.from_numpy(weights_data[weights_index:(weights_index + bias_size)])
                    bias_data = bias_data.view_as(conv2d.bias.data)
                    conv2d.bias.data.copy_(bias_data)
                    weights_index += bias_size

                # Load weight data for convolutional module
                weight_size = conv2d.weight.numel()
                weight_data = torch.from_numpy(weights_data[weights_index:(weights_index + weight_size)])
                weight_data = weight_data.view_as(conv2d.weight.data)
                conv2d.weight.data.copy_(weight_data)
                weights_index += weight_size

                # Add convolutional module
                module.add_module("conv2d_" + str(layer_index), conv2d)

                # Add batch normalization module
                if batch_normalize:
                    module.add_module("batchnorm2d_" + str(layer_index), batchnorm2d)

                # Create and add activation module
                if activation == "leaky":
                    leakyrelu = nn.LeakyReLU(header.leakyrelu_negative_slope, inplace = True)
                    module.add_module("leakyrelu_" + str(layer_index), leakyrelu)

                # Store next input size
                input_channel_size_next = filters
            # Process upsample configurations
            elif configuration["type"] == "upsample":
                # Read stride
                stride = int(configuration["stride"])
                upsample = nn.Upsample(scale_factor = stride, mode = header.upsample_mode, align_corners = True)

                # Create and add upsample module
                module.add_module("upsample_" + str(layer_index), upsample)
            # Process route configurations
            elif configuration["type"] == "route":
                # Read route indices
                route_indices = configuration["layers"].split(',')
                route_indices = [int(route_index) for route_index in route_indices] # Convert to integer
                route_index_first = route_indices[0]
                route_index_second = 0

                # Read second route index if exists
                if (len(route_indices) > 1):
                    route_index_second = route_indices[1]

                # Convert first index from relative to absolute
                if route_index_first < 0:
                    route_index_first += layer_index

                # Convert second index from relative to absolute
                if route_index_second < 0:
                    route_index_second += layer_index

                # If second route index exists
                if route_index_second != 0:
                    # Compute next input channel size
                    input_channel_size_next = input_channel_sizes[route_index_first - 1] + input_channel_sizes[route_index_second - 1]

                    # Flag layer output to be cached
                    layer_output_caching_flags[route_index_first] = True
                    layer_output_caching_flags[route_index_second] = True
                # If second route index does not exist
                else:
                    # Compute next input channel size
                    input_channel_size_next = input_channel_sizes[route_index_first - 1]

                    # Flag layer output to be cached
                    layer_output_caching_flags[route_index_first] = True

                # Create and add route module
                module.add_module("route_" + str(layer_index), nn.Module())
            # Process shortcut configurations
            elif configuration["type"] == "shortcut":
                # Read shortcut index
                shortcut_index = configuration["from"].split(',')
                shortcut_index = int(shortcut_index[0])   # Convert to integer

                # Convert shortcut index from relative to absolute
                if shortcut_index < 0:
                    shortcut_index += layer_index

                # Flag layer output to be cached
                layer_output_caching_flags[shortcut_index] = True

                # Create and add shortcut module
                module.add_module("shortcut_" + str(layer_index), nn.Module())
            # Process YOLO configurations
            elif configuration["type"] == "yolo":
                # Create and add YOLO module
                module.add_module("yolo_" + str(layer_index), nn.Module())
            # Process unknown configurations
            else:
                # Log warning message
                logger.log_warn("Unknown layer type.")

            # Append current module
            module_list.append(module)
            input_channel_sizes.append(input_channel_size_next)

        # Return modules
        return network_information, weights_header, module_list, layer_output_caching_flags

    def inference(self, input):
        # Declare output and layer output variables
        output = []
        layer_output = input
        layer_output_cache = {}

        # Process modules
        for layer_index, configuration in enumerate(self.configurations[1:]):
            # Process convolutional module
            if configuration["type"] == "convolutional":
                module = self.module_list[layer_index]
                layer_output = module(layer_output)
            # Process upsample module
            elif configuration["type"] == "upsample":
                module = self.module_list[layer_index]
                layer_output = module(layer_output)
            # Process route module
            elif configuration["type"] == "route":
                # Read route indices
                route_indices = configuration["layers"].split(',')
                route_indices = [int(route_index) for route_index in route_indices] # Convert to integer
                route_index_first = route_indices[0]
                route_index_second = 0

                # Read second route index if exists
                if (len(route_indices) > 1):
                    route_index_second = route_indices[1]

                # Convert first index from relative to absolute
                if route_index_first < 0:
                    route_index_first += layer_index

                # Convert second index from relative to absolute
                if route_index_second < 0:
                    route_index_second += layer_index

                # If second route index exists
                if route_index_second != 0:
                    # Read from cached outputs
                    layer_output_first = layer_output_cache[route_index_first]
                    layer_output_second = layer_output_cache[route_index_second]

                    # Concatenate cached layer outputs
                    layer_output = torch.cat((layer_output_first, layer_output_second), 1)
                # If second route index does not exist
                else:
                    # Read from cached layer output
                    layer_output = layer_output_cache[route_index_first]
            # Process shortcut module
            elif configuration["type"] == "shortcut":
                # Read shortcut index
                shortcut_index = configuration["from"].split(',')
                shortcut_index = int(shortcut_index[0])   # Convert to integer

                # Convert shortcut index from relative to absolute
                if shortcut_index < 0:
                    shortcut_index += layer_index

                # Add shortcut layer output to last layer output
                layer_output += layer_output_cache[shortcut_index]
            # Process YOLO module
            elif configuration["type"] == "yolo":
                # Store last layer output and its index
                output.append((layer_index, layer_output))
            # Process unknown module
            else:
                # Log warning message
                logger.log_warn("Unknown layer type.")

            # Cache layer output
            if self.layer_output_caching_flags[layer_index]:
                layer_output_cache[layer_index] = layer_output

            # Update component timers
            if header.time_components:
                utility.update_component_timers(layer_index)

        # Return output
        return output
