#! /usr/bin/env python3

import argparse
import copy
import cv2
import darknet
import difference
import evaluation
import header
import json
import logger
import matplotlib.pyplot
import os
import pycocotools.coco
import query
import sys
import time
import torch
import utility

def main():
    # Initialize environment
    utility.initialize_environment()

    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--draw-bounding-box", "-b", type = str, help = "Whether to draw bounding box.", required = False)
    parser.add_argument("--focusification-mask", "-m", type = str, help = "Path to focusification mask image file.", required = False)
    parser.add_argument("--focusification-mask-pt", "-p", type = str, help = "Path to focusification mask PT file.", required = False)
    parser.add_argument("--focusify", "-f", type = str, help = "Whether to focusify model.", required = False)
    parser.add_argument("--input", "-i", type = str, help = "Path to input image directory.", required = False)
    parser.add_argument("--input-size", "-s", type = str, help = "Size of the network input layer.", required = False)
    parser.add_argument("--difference-threshold", "-d", type = str, help = "Difference detection threshold.", required = False)
    args = parser.parse_args()

    # Process draw bounding box arguemnt
    if args.draw_bounding_box:
        try:
            if int(args.draw_bounding_box) != 0:
                header.draw_bounding_box = True
            else:
                header.draw_bounding_box = False
        except:
            pass

    # Process focusification mask arguemnt
    if args.focusification_mask:
        header.file_path_name_focusification_mask = args.focusification_mask

    # Process focusification mask PT arguemnt
    if args.focusification_mask_pt:
        header.file_path_name_focusification_mask_pt = args.focusification_mask_pt
        header.focusification_mask = torch.load(header.file_path_name_focusification_mask_pt)

    # Process focusify arguemnt
    if args.focusify:
        try:
            if int(args.focusify) != 0:
                header.focusify = True
            else:
                header.focusify = False
        except:
            pass

    # Process input arguemnt
    if args.input:
        header.file_path_name_input = args.input

    # Validate input image directory
    if not os.path.isdir(header.file_path_name_input):
        logger.log_fatal("Invalid input image directory.")
        exit(1)

    # Process input size arguemnt
    if args.input_size:
        if int(args.input_size) % header.network_stride:
            logger.log_warn("Invalid input size: " + args.input_size + ".")
        else:
            header.input_size = int(args.input_size)

    # Process difference detection threshold arguemnt
    if args.difference_threshold:
        header.threshold_difference_detection = float(args.difference_threshold)

    # Open timing log file
    os.makedirs(header.log_path, exist_ok = True)
    header.log_file_timing = open(header.log_path + "/" + "timing_" + str(header.input_size) + ".log", "w+")

    # Open result log file
    if not os.path.isfile(os.path.join(header.log_path, "result.log")):
        header.log_file_result = open(os.path.join(header.log_path, "result.log"), "w")
        header.log_file_result.write("#\tinput\tfocusification_mask\tfocusification_mask_pt\tfocusify\tthreshold_difference_detection\ttiming_overall_inference\ttiming_overall_query_aggregate\ttiming_overall_query_limit\tmean_average_precision_overall\tquery_aggregate\tquery_limit\n")
        header.log_file_result.flush()
    else:
        header.log_file_result = open(os.path.join(header.log_path, "result.log"), "a")

    # Open ground truth file
    sys.stdout = open(os.devnull, 'w')
    file_ground_truth = pycocotools.coco.COCO(header.file_path_name_ground_truth)
    sys.stdout = sys.__stdout__

    # Create formatted ground truth file
    utility.format_gt_for_mAP(file_ground_truth)

    # Open formatted ground truth file
    with open(header.file_path_name_ground_truth_formatted) as file_ground_truth_formatted_raw:
        file_ground_truth_formatted = json.load(file_ground_truth_formatted_raw)

    # Load focusification mask
    logger.log_info("Loading focusification mask \"" + header.file_path_name_focusification_mask + "\"...")
    focusification_mask = utility.load_focusification_mask(header.file_path_name_focusification_mask)

    # Use focusification mask
    if header.focusify and header.use_focusification_mask:
        logger.log_info("Using focusification mask \"" + header.file_path_name_focusification_mask + "\"...")
        header.focusification_mask = focusification_mask

    logger.log_debug("Focusification mask shape:", header.focusification_mask.shape)

    # Draw focusification mask
    if header.draw_focusification_mask:
        matplotlib.pyplot.imshow(header.focusification_mask.numpy())
        matplotlib.pyplot.show()

    # Create network
    if header.focusify:
        logger.log_info("Creating focusified network...")
    else:
        logger.log_info("Creating network...")

    network = darknet.Darknet(header.file_path_name_config, header.file_path_name_weights)

    # Set evaluation mode
    if header.evaluation_mode:
        logger.log_info("Setting evaluation mode...")
        network.eval()

    # Initialize CUDA
    if header.cuda_mode:
        if header.cuda_available:
            logger.log_info("Initializing CUDA...")
            network.cuda()
        else:
            logger.log_info("CUDA not available.")

    # Declare variables
    input_original_last = None
    bounding_boxes_last = None
    mean_average_precision_list = []
    queries_aggregate = []
    queries_limit = []
    timings_inference = []
    timings_inference_query_limit = []
    timings_query_aggregate = []
    timings_query_limit = []

    # Perform inference
    for file_name_input in sorted(os.listdir(header.file_path_name_input)):
        # Split input file name
        file_name_input_split_postfix = file_name_input.split(".")
        file_name_input_split_prefix = file_name_input.split(header.file_name_delimiter_input)

        # Filter input file name by postfix
        if len(file_name_input_split_postfix) < 2 or file_name_input_split_postfix[1] != header.file_postfix_input:
            continue

        # Filter input file name by prefix
        if len(file_name_input_split_prefix) < 2 or file_name_input_split_prefix[0] != header.file_prefix_input:
            continue

        # Read input
        logger.log_trace("Reading input file \"" + header.file_path_name_input + "\"...")
        input_original = cv2.imread(os.path.join(header.file_path_name_input, file_name_input))

        # Preprocess input
        logger.log_trace("Preprocessing input at size " + str(header.input_size) + "...")
        input = utility.preprocess(input_original)

        # Initialize CUDA
        if header.cuda_mode and header.cuda_available:
            input = input.cuda()

        if input_original_last is None or difference.detect_difference(input_original, input_original_last, focusification_mask):
            # Update start-of-neck timer
            header.timer_start_of_neck = time.time() * 1000

            # Perform inference
            output = network.inference(input)

            # Wait for CUDA to finish
            if header.cuda_mode and header.cuda_available:
                torch.cuda.synchronize()

            # Update end-of-head timer
            header.timer_end_of_head = time.time() * 1000

            # Record timings
            utility.record_timings()

            # Read configurations
            configurations = network.get_configurations()

            # Read small scale output
            output_scale_small_layer_index = output[0][0]
            output_scale_small = output[0][1]
            output_scale_small_configuration = configurations[output_scale_small_layer_index + 1]

            # Read medium scale output
            output_scale_medium_layer_index = output[1][0]
            output_scale_medium = output[1][1]
            output_scale_medium_configuration = configurations[output_scale_medium_layer_index + 1]

            # Read large scale output
            output_scale_large_layer_index = output[2][0]
            output_scale_large = output[2][1]
            output_scale_large_configuration = configurations[output_scale_large_layer_index + 1]

            # Postprocess small scale output
            logger.log_trace("Postprocessing small scale output...")
            bounding_boxes_scale_small = utility.postprocess(input_original, output_scale_small, output_scale_small_configuration)

            # Postprocess medium scale output
            logger.log_trace("Postprocessing medium scale output...")
            bounding_boxes_scale_medium = utility.postprocess(input_original, output_scale_medium, output_scale_medium_configuration)

            # Postprocess large scale output
            logger.log_trace("Postprocessing large scale output...")
            bounding_boxes_scale_large = utility.postprocess(input_original, output_scale_large, output_scale_large_configuration)

            # Declare concatenated bounding box list
            bounding_boxes = [[] for _ in range(header.num_classes)]

            # Concatenate outputs from all scales
            for class_index in range(0, header.num_classes):
                bounding_boxes[class_index] += bounding_boxes_scale_small[class_index]
                bounding_boxes[class_index] += bounding_boxes_scale_medium[class_index]
                bounding_boxes[class_index] += bounding_boxes_scale_large[class_index]

            # Perform non-maximum_suppression
            bounding_boxes = utility.perform_nms(bounding_boxes)

            # Calculate mean average precision
            (mean_average_precision, ground_truth_boxes) = evaluation.format_for_mAP(bounding_boxes, file_name_input, file_ground_truth, file_ground_truth_formatted, header.query_class)

            # Log results
            logger.log_info(str(len(bounding_boxes)) + " bounding box(es) predicted for \"" + file_name_input + "\" in " + "{:.2f}".format(header.timing_total) + " ms with mAP of " + str(round(mean_average_precision, 2)) + ".")

            # Store inference timing
            timings_inference.append(header.timing_total)

            # Store limit query inference timing
            if len(queries_limit) < header.query_limit:
                timings_inference_query_limit.append(header.timing_total)

            # Store last input and output
            input_original_last = copy.deepcopy(input_original)
            bounding_boxes_last = copy.deepcopy(bounding_boxes)
        else:
            # Load from last input and output
            logger.log_info("Skipping frame \"" + file_name_input + "\" with mAP of " + str(round(mean_average_precision, 2)) + ".")
            bounding_boxes = copy.deepcopy(bounding_boxes_last)

            # Calculate mean average precision
            (mean_average_precision, ground_truth_boxes) = evaluation.format_for_mAP(bounding_boxes, file_name_input, file_ground_truth, file_ground_truth_formatted, header.query_class)

            # Store inference timing
            timings_inference.append(0)

            # Store limit query inference timing
            if len(queries_limit) < header.query_limit:
                timings_inference_query_limit.append(0)

        # Store mean average precision
        mean_average_precision_list.append(mean_average_precision)

        # Update aggregate queries
        header.timer_start_of_query_aggregate = time.time() * 1000
        queries_aggregate.append(query.query_aggregate(bounding_boxes, header.query_class))
        header.timer_end_of_query_aggregate = time.time() * 1000

        # Store aggregate query timing
        timings_query_aggregate.append(header.timer_end_of_query_aggregate - header.timer_start_of_query_aggregate)

        # Update limit queries
        if len(queries_limit) < header.query_limit:
            header.timer_start_of_query_limit = time.time() * 1000
            if query.query_limit(bounding_boxes, header.query_class, header.query_limit_threshold, header.query_limit_threshold_above):
                queries_limit.append(file_name_input)
            header.timer_end_of_query_limit = time.time() * 1000

             # Store limit query timing
            timings_query_limit.append(header.timer_end_of_query_limit - header.timer_start_of_query_limit)

        # Draw bounding boxes and display input
        if header.draw_bounding_box:
            # Draw bounding boxes
            logger.log_trace("Drawing bounding boxes...")
            utility.draw_bounding_box(input_original, bounding_boxes, ground_truth_boxes)

            # Display input with bounding boxes drawn
            logger.log_trace("Displaying input with bounding boxes drawn...")
            cv2.imshow(header.file_path_name_input, input_original)
            cv2.waitKey(1)

    # Calculate overall timings
    timing_overall_inference = str(round(sum(timings_inference), 2))
    timing_overall_query_aggregate = str(round(sum(timings_inference) + sum(timings_query_aggregate), 2))
    timing_overall_query_limit = str(round(sum(timings_inference_query_limit) + sum(timings_query_limit), 2))

    # Log overall timings
    logger.log_info("Overall inference time: " + timing_overall_inference + " ms.")
    logger.log_info("Overall aggregate query time: " + timing_overall_query_aggregate + " ms.")
    logger.log_info("Overall limit query time: " + timing_overall_query_limit + " ms.")

    # Calculate overall mAP
    if len(mean_average_precision_list) != 0:
        mean_average_precision_overall = str(round(sum(mean_average_precision_list) / len(mean_average_precision_list), 2))
    else:
        mean_average_precision_overall = 0

    # Log overall mAP
    logger.log_info("Overall inference mAP: " + mean_average_precision_overall + ".")

    # Load query class name
    classes = utility.load_classes(header.file_path_name_class)
    query_class_name = classes[header.query_class]

    # Calculate aggregate query result
    if len(queries_aggregate) != 0:
        query_aggregate = str(int(sum(queries_aggregate) / len(queries_aggregate)))
    else:
        query_aggregate = 0

    # Log aggregate query result
    logger.log_info("Aggregate query (average number of \"" + query_class_name + "\"): " + query_aggregate + ".")

    # Calculate limit query result
    query_limit = str(queries_limit)

    # Log limit query result
    if header.query_limit_threshold_above:
        logger.log_info("Limit query (first " + str(header.query_limit) + " frames with number of \"" + query_class_name + "\" above " + str(header.query_limit_threshold) + "): " + query_limit + ".")
    else:
        logger.log_info("Limit query (first " + str(header.query_limit) + " frames with number of \"" + query_class_name + "\" equal or below " + str(header.query_limit_threshold) + "): " + query_limit + ".")

    # Log all results
    log_result_line = header.file_path_name_input
    log_result_line += '\t'
    log_result_line += header.file_path_name_focusification_mask
    log_result_line += '\t'
    log_result_line += header.file_path_name_focusification_mask_pt
    log_result_line += '\t'
    log_result_line += str(header.focusify)
    log_result_line += '\t'
    log_result_line += str(header.threshold_difference_detection)
    log_result_line += '\t'
    log_result_line += timing_overall_inference
    log_result_line += '\t'
    log_result_line += timing_overall_query_aggregate
    log_result_line += '\t'
    log_result_line += timing_overall_query_limit
    log_result_line += '\t'
    log_result_line += mean_average_precision_overall
    log_result_line += '\t'
    log_result_line += query_aggregate
    log_result_line += '\t'
    log_result_line += query_limit
    log_result_line += '\n'
    header.log_file_result.write(log_result_line)

    # Close log files
    header.log_file_result.close()
    header.log_file_timing.close()

    return

if __name__ == "__main__":
    main()
