import cv2
import gc
import header
import json
import logger
import math
import numpy
import os
import PIL.Image
import time as tm
import torch
import torch.autograd
import torchvision

def initialize_environment():
    logger.log_info("Initializing environment...")

    # Perform garbage collection
    gc.collect()

    # Initialize CUDA
    if header.cuda_mode and header.cuda_available:
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.empty_cache()
        torch.cuda.synchronize()

    return

def load_classes(file_path_name):
    # Open class file
    file = open(file_path_name, "r")

    # Read lines
    lines = file.readlines()
    lines = [line for line in lines if line[0] != '#']  # Remove comments
    lines = [line.strip() for line in lines]    # Remove white spaces
    lines = [line for line in lines if line != ""]  # Remove empty lines

    # Declare class list
    classes = []

    # Process lines
    for line in lines:
        # Store class name to class list
        classes.append(line)

    # Return class list
    return classes

def load_focusification_mask(file_path_name):
    focusification_mask = PIL.Image.open(file_path_name).convert('1')
    focusification_mask = focusification_mask.resize((header.input_size, header.input_size))
    focusification_mask = torchvision.transforms.functional.pil_to_tensor(focusification_mask)
    focusification_mask = torch.squeeze(focusification_mask, 0)

    return focusification_mask

def preprocess(input):
    input = cv2.resize(input, (header.input_size, header.input_size))   # Resize to match network input dimension
    input = input[:, :, ::-1].transpose((2, 0, 1))  # Convert BGR to RGB and H x W x C to C x H x W
    input = input[numpy.newaxis, :, :, :]  # Add batch channel
    input = input / header.input_normalization  # Normalize input
    input = torch.from_numpy(input).float() # Convert to float
    input = torch.autograd.Variable(input)  # Convert to autograd variable

    return input

def time():
    # Wait for CUDA to finish
    if header.cuda_mode and header.cuda_available:
        torch.cuda.synchronize()

    # Return current time
    return tm.time()

def time_component():
    if not header.time_components:
        return 0

    # Return current time
    return time()

def update_component_timers(layer_index):
    logger.log_debug("Updating component timers...")

    # Wait for CUDA to finish
    if header.cuda_mode and header.cuda_available:
        torch.cuda.synchronize()

    # Update end-of-neck timer
    if layer_index == header.layer_index_end_of_neck:
        header.timer_end_of_neck = tm.time() * 1000

    # Update start-of-head timer
    if layer_index == header.layer_index_start_of_head:
        header.timer_start_of_head = tm.time() * 1000

    # Update start-of-small-scale timer
    if layer_index == header.layer_index_start_of_scale_small:
        header.timer_start_of_scale_small = tm.time() * 1000

    # Update start-of-small-scale-tail timer
    if layer_index == header.layer_index_start_of_scale_tail_small:
        header.timer_start_of_scale_tail_small = tm.time() * 1000

    # Update end-of-small-scale-tail timer
    if layer_index == header.layer_index_end_of_scale_tail_small:
        header.timer_end_of_scale_tail_small = tm.time() * 1000

    # Update end-of-small-scale timer
    if layer_index == header.layer_index_end_of_scale_small:
        header.timer_end_of_scale_small = tm.time() * 1000

    # Update start-of-medium-scale timer
    if layer_index == header.layer_index_start_of_scale_medium:
        header.timer_start_of_scale_medium = tm.time() * 1000

    # Update start-of-medium-scale-tail timer
    if layer_index == header.layer_index_start_of_scale_tail_medium:
        header.timer_start_of_scale_tail_medium = tm.time() * 1000

    # Update end-of-medium-scale-tail timer
    if layer_index == header.layer_index_end_of_scale_tail_medium:
        header.timer_end_of_scale_tail_medium = tm.time() * 1000

    # Update end-of-medium-scale timer
    if layer_index == header.layer_index_end_of_scale_medium:
        header.timer_end_of_scale_medium = tm.time() * 1000

    # Update start-of-large-scale timer
    if layer_index == header.layer_index_start_of_scale_large:
        header.timer_start_of_scale_large = tm.time() * 1000

    # Update start-of-large-scale-tail timer
    if layer_index == header.layer_index_start_of_scale_tail_large:
        header.timer_start_of_scale_tail_large = tm.time() * 1000

    # Update end-of-large-scale-tail timer
    if layer_index == header.layer_index_end_of_scale_tail_large:
        header.timer_end_of_scale_tail_large = tm.time() * 1000

    # Update end-of-large-scale timer
    if layer_index == header.layer_index_end_of_scale_large:
        header.timer_end_of_scale_large = tm.time() * 1000

    return

def record_timings():
    # Calculate total timing
    header.timing_total = header.timer_end_of_head - header.timer_start_of_neck

    # Calculate and log component timings and timing percentages
    if header.time_components:
        # Calculate component timings
        timing_neck = header.timer_end_of_neck - header.timer_start_of_neck
        timing_scale_tail_small = header.timer_end_of_scale_tail_small - header.timer_start_of_scale_tail_small
        timing_scale_small = header.timer_end_of_scale_small - header.timer_start_of_scale_small
        timing_scale_tail_medium = header.timer_end_of_scale_tail_medium - header.timer_start_of_scale_tail_medium
        timing_scale_medium = header.timer_end_of_scale_medium - header.timer_start_of_scale_medium
        timing_scale_tail_large = header.timer_end_of_scale_tail_large - header.timer_start_of_scale_tail_large
        timing_scale_large = header.timer_end_of_scale_large - header.timer_start_of_scale_large
        timing_head = header.timer_end_of_head - header.timer_start_of_head

        # Calculate component timing percentages
        timing_percentage_neck = timing_neck / header.timing_total * 100
        timing_percentage_scale_tail_small = timing_scale_tail_small / header.timing_total * 100
        timing_percentage_scale_small = timing_scale_small / header.timing_total * 100
        timing_percentage_scale_tail_medium = timing_scale_tail_medium / header.timing_total * 100
        timing_percentage_scale_medium = timing_scale_medium / header.timing_total * 100
        timing_percentage_scale_tail_large = timing_scale_tail_large / header.timing_total * 100
        timing_percentage_scale_large = timing_scale_large / header.timing_total * 100
        timing_percentage_head = timing_head / header.timing_total * 100

        # Log component timing percentages
        logger.log_info("Neck inference took " + "{:.2f}".format(timing_percentage_neck) + "% of the time.")
        logger.log_info("Small scale tail inference took " + "{:.2f}".format(timing_percentage_scale_tail_small) + "% of the time.")
        logger.log_info("Small scale inference took " + "{:.2f}".format(timing_percentage_scale_small) + "% of the time.")
        logger.log_info("Medium scale tail inference took " + "{:.2f}".format(timing_percentage_scale_tail_medium) + "% of the time.")
        logger.log_info("Medium scale inference took " + "{:.2f}".format(timing_percentage_scale_medium) + "% of the time.")
        logger.log_info("Large scale tail inference took " + "{:.2f}".format(timing_percentage_scale_tail_large) + "% of the time.")
        logger.log_info("Large scale inference took " + "{:.2f}".format(timing_percentage_scale_large) + "% of the time.")
        logger.log_info("Head inference took " + "{:.2f}".format(timing_percentage_head) + "% of the time.")

        # Log component timings
        logger.log_info("Neck inference took " + "{:.2f}".format(timing_neck) + " ms.")
        logger.log_info("Small scale tail inference took " + "{:.2f}".format(timing_scale_tail_small) + " ms.")
        logger.log_info("Small scale inference took " + "{:.2f}".format(timing_scale_small) + " ms.")
        logger.log_info("Medium scale tail inference took " + "{:.2f}".format(timing_scale_tail_medium) + " ms.")
        logger.log_info("Medium scale inference took " + "{:.2f}".format(timing_scale_medium) + " ms.")
        logger.log_info("Large scale tail inference took " + "{:.2f}".format(timing_scale_tail_large) + " ms.")
        logger.log_info("Large scale inference took " + "{:.2f}".format(timing_scale_large) + " ms.")
        logger.log_info("Head inference took " + "{:.2f}".format(timing_head) + " ms.")

    # Log total timing
    header.log_file_timing.write(str(header.timing_total) + "\n")

    return

def calculate_diou_penalty(bounding_box_1, bounding_box_2):
    # https://arxiv.org/pdf/1911.08287.pdf

    # Read bounding box 1 diagonal points
    bounding_box_1_upper_left_x = bounding_box_1[0]
    bounding_box_1_upper_left_y = bounding_box_1[1]
    bounding_box_1_lower_right_x = bounding_box_1[2]
    bounding_box_1_lower_right_y = bounding_box_1[3]

    # Read bounding box 1 diagonal points
    bounding_box_2_upper_left_x = bounding_box_2[0]
    bounding_box_2_upper_left_y = bounding_box_2[1]
    bounding_box_2_lower_right_x = bounding_box_2[2]
    bounding_box_2_lower_right_y = bounding_box_2[3]

    # Calculate diagonal points of the smallest enclosing box covering the two bounding boxes
    bounding_box_enclosing_upper_left_x = min(bounding_box_1_upper_left_x, bounding_box_2_upper_left_x)
    bounding_box_enclosing_upper_left_y = min(bounding_box_1_upper_left_y, bounding_box_2_upper_left_y)
    bounding_box_enclosing_lower_right_x = max(bounding_box_1_lower_right_x, bounding_box_2_lower_right_x)
    bounding_box_enclosing_lower_right_y = max(bounding_box_1_lower_right_y, bounding_box_2_lower_right_y)

    # Calculate bounding box dimensions
    bounding_box_1_width = max(0, bounding_box_1_lower_right_x - bounding_box_1_upper_left_x)
    bounding_box_1_height = max(0, bounding_box_1_lower_right_y - bounding_box_1_upper_left_y)
    bounding_box_2_width = max(0, bounding_box_2_lower_right_x, bounding_box_2_upper_left_x)
    bounding_box_2_height = max(0, bounding_box_2_lower_right_y, bounding_box_2_upper_left_y)
    bounding_box_enclosing_width = max(0, bounding_box_enclosing_lower_right_x - bounding_box_enclosing_upper_left_x)
    bounding_box_enclosing_height = max(0, bounding_box_enclosing_lower_right_y - bounding_box_enclosing_upper_left_y)

    # Calculate bounding box center points
    bounding_box_1_center_x = bounding_box_2_upper_left_x + bounding_box_1_width / 2
    bounding_box_1_center_y = bounding_box_2_upper_left_y + bounding_box_1_height / 2
    bounding_box_2_center_x = bounding_box_2_upper_left_x + bounding_box_2_width / 2
    bounding_box_2_center_y = bounding_box_2_upper_left_y + bounding_box_2_height / 2

    # Calculate rho, Euclidean distance between the two bounding box center points
    rho = math.sqrt((bounding_box_2_center_x - bounding_box_1_center_x) ** 2 +
        (bounding_box_2_center_y - bounding_box_1_center_y) ** 2)

    # Calculate c, diagonal length of the smallest enclosing box covering the two bounding boxes
    c = math.sqrt((bounding_box_enclosing_width) ** 2 + (bounding_box_enclosing_height) ** 2)

    return (rho ** 2) / (c ** 2)

def calculate_iou(bounding_box_1, bounding_box_2):
    # Declare DIoU penalty
    diou_penalty = 0

    # Read bounding box 1 diagonal points
    bounding_box_1_upper_left_x = bounding_box_1[0]
    bounding_box_1_upper_left_y = bounding_box_1[1]
    bounding_box_1_lower_right_x = bounding_box_1[2]
    bounding_box_1_lower_right_y = bounding_box_1[3]

    # Read bounding box 1 diagonal points
    bounding_box_2_upper_left_x = bounding_box_2[0]
    bounding_box_2_upper_left_y = bounding_box_2[1]
    bounding_box_2_lower_right_x = bounding_box_2[2]
    bounding_box_2_lower_right_y = bounding_box_2[3]

    # Calculate bounding box dimensions
    bounding_box_1_width = max(0, bounding_box_1_lower_right_x - bounding_box_1_upper_left_x)
    bounding_box_1_height = max(0, bounding_box_1_lower_right_y - bounding_box_1_upper_left_y)
    bounding_box_2_width = max(0, bounding_box_2_lower_right_x - bounding_box_2_upper_left_x)
    bounding_box_2_height = max(0, bounding_box_2_lower_right_y - bounding_box_2_upper_left_y)

    # Calculate bounding box areas
    bounding_box_1_area = bounding_box_1_width * bounding_box_1_height
    bounding_box_2_area = bounding_box_2_width * bounding_box_2_height

    # Calculate intersection
    intersection_width = max(0, min(bounding_box_1_lower_right_x, bounding_box_2_lower_right_x) -
        max(bounding_box_1_upper_left_x, bounding_box_2_upper_left_x))
    intersection_height = max(0, min(bounding_box_1_lower_right_y, bounding_box_2_lower_right_y) -
        max(bounding_box_1_upper_left_y, bounding_box_2_upper_left_y))
    intersection = intersection_width * intersection_height

    # Calculate union
    union = bounding_box_1_area + bounding_box_2_area - intersection

    # Calculate DIoU penalty
    if header.use_diou:
        diou_penalty = calculate_diou_penalty(bounding_box_1, bounding_box_2)

    # Return intersection over union
    return (intersection / union) - diou_penalty

def postprocess(input_original, output, configuration):
    # Declare bounding box list
    bounding_boxes = [[] for _ in range(header.num_classes)]

    # Validate configuration
    if configuration["type"] != "yolo":
        logger.log_error("Incorrect layer configuration type: " + configuration["type"] + ". Expected \"yolo\".")
        return bounding_boxes

    # Read anchor mask
    anchor_mask = configuration["mask"].split(',')
    anchor_mask = [int(index) for index in anchor_mask] # Convert to integer

    # Read anchors
    anchors = configuration["anchors"].split(',')
    anchors = [int(anchor) for anchor in anchors]   # Convert to integer
    anchors = [(anchors[i], anchors[i + 1]) for i in range(0, len(anchors), 2)] # Store in pairs

    # Mask anchors using anchor mask
    anchors = [anchors[i] for i in anchor_mask]

    # Read and compute output dimensions
    batch_size = output.size(0)
    grid_scale = output.size(2)
    prediction_size = 5 + header.num_classes    # 5 = 4 bounding box parameters + 1 objectness score

    # View output as (batch_size) x (prediction_size x num_anchors) x (grid_scale x grid_scale)
    output = output.view(batch_size, prediction_size * header.num_anchors, grid_scale * grid_scale)

    # Convert output to (batch_size) x (grid_scale x grid_scale) x (prediction_size x num_anchors)
    output = output.transpose(1, 2).contiguous()

    # View output as (batch_size) x (grid_scale x grid_scale x num_anchors) x (prediction_size)
    output = output.view(batch_size, grid_scale * grid_scale * header.num_anchors, prediction_size)

    # Calculate input resize factors
    input_resize_factor_x = input_original.shape[1] / header.input_size
    input_resize_factor_y = input_original.shape[0] / header.input_size

    # Process output
    for batch_index in range(0, batch_size):
        for bounding_box_index in range(0, grid_scale * grid_scale * header.num_anchors):
            # Read objectness score
            objectness_score = torch.sigmoid(output[batch_index][bounding_box_index][4])

            # Threshold objectness score
            if objectness_score < header.threshold_objectness_score:
                continue

            # Declare bounding box parameter list
            bounding_box = []

            # Compute anchor index and grid parameters
            anchor_index = bounding_box_index % header.num_anchors
            grid_size = header.input_size // grid_scale
            grid_index = bounding_box_index // header.num_anchors
            grid_index_col = grid_index % grid_scale
            grid_index_row = grid_index // grid_scale

            # Read bounding box parameters
            bounding_box_tx = torch.sigmoid(output[batch_index][bounding_box_index][0])
            bounding_box_ty = torch.sigmoid(output[batch_index][bounding_box_index][1])
            bounding_box_tw = output[batch_index][bounding_box_index][2]
            bounding_box_th = output[batch_index][bounding_box_index][3]

            # Convert bounding box parameters to input pixel coordinates
            bounding_box_x = (grid_index_col + bounding_box_tx) * grid_size
            bounding_box_y = (grid_index_row + bounding_box_ty) * grid_size
            bounding_box_w = anchors[anchor_index][0] * math.exp(bounding_box_tw)
            bounding_box_h = anchors[anchor_index][1] * math.exp(bounding_box_th)

            # Convert bounding box parameters to diagonal points
            bounding_box_upper_left_x = bounding_box_x - bounding_box_w / 2
            bounding_box_upper_left_y = bounding_box_y - bounding_box_h / 2
            bounding_box_lower_right_x = bounding_box_x + bounding_box_w / 2
            bounding_box_lower_right_y = bounding_box_y + bounding_box_h / 2

            # Store bounding box parameters
            bounding_box.append(int(bounding_box_upper_left_x * input_resize_factor_x))
            bounding_box.append(int(bounding_box_upper_left_y * input_resize_factor_y))
            bounding_box.append(int(bounding_box_lower_right_x * input_resize_factor_x))
            bounding_box.append(int(bounding_box_lower_right_y * input_resize_factor_y))
            bounding_box.append(float(objectness_score))

            # Declare max class score variables
            class_score_max = -float("inf")
            class_index = 0

            # Read class scores
            for class_score_index in range(5, prediction_size):
                class_score = torch.sigmoid(output[batch_index][bounding_box_index][class_score_index])

                # Check max class score
                if class_score > class_score_max:
                    class_score_max = class_score
                    class_index = class_score_index - 5

            # Threshold class score
            if class_score_max < header.threshold_class_score:
                continue

            # Store class score and class index
            bounding_box.append(float(class_score_max))
            bounding_box.append(int(class_index))

            # Store bounding box
            bounding_boxes[class_index].append(bounding_box)

    # Return bounding boxes
    return bounding_boxes

def perform_nms(bounding_boxes_classes):
    # Declare bounding box list
    bounding_boxes = []

    # Perform non-maximum suppression
    for class_index in range(0, header.num_classes):
        # Read list of bounding boxes with the same class
        bounding_boxes_class = bounding_boxes_classes[class_index]

        # Declare bounding box class list flags
        bounding_boxes_class_flag = [True] * len(bounding_boxes_class)

        # Process first bounding box
        for bounding_box_1_index in range(0, len(bounding_boxes_class)):
            # Read first bounding box
            bounding_box_1 = bounding_boxes_class[bounding_box_1_index]

            # Process second bounding box
            for bounding_box_2_index in range(0, len(bounding_boxes_class)):
                # Skip if bounding box 2 is the same as bounding box 1
                if bounding_box_2_index == bounding_box_1_index:
                    continue

                # Read bounding box 2 and calculate IoU
                bounding_box_2 = bounding_boxes_class[bounding_box_2_index]
                iou = calculate_iou(bounding_box_1, bounding_box_2)

                # Check if IoU is greater than threshold, i.e., if they are considered the same object
                if iou > header.threshold_non_maximum_suppression:
                    # Read class scores
                    bounding_box_1_class_score = bounding_box_1[5]
                    bounding_box_2_class_score = bounding_box_2[5]

                    # Mark bounding box 1 for removal
                    if (bounding_box_1_class_score <= bounding_box_2_class_score and
                        bounding_boxes_class_flag[bounding_box_1_index]):
                        bounding_boxes_class_flag[bounding_box_1_index] = False

        # Process bounding box class list
        for bounding_box_index in range(0, len(bounding_boxes_class)):
            if bounding_boxes_class_flag[bounding_box_index]:
                bounding_boxes.append(bounding_boxes_class[bounding_box_index])

    # Return bounding boxes
    return bounding_boxes

def draw_bounding_box(input, bounding_boxes, ground_truth_boxes):
    # Load COCO classes
    classes = load_classes(header.file_path_name_class)

    # Process ground truth boxes
    for i in range(0, len(ground_truth_boxes["boxes"])):
        # Read ground truth box parameters
        ground_truth_upper_left_x = ground_truth_boxes["boxes"][i][0]
        ground_truth_upper_left_y = ground_truth_boxes["boxes"][i][1]
        ground_truth_lower_right_x = ground_truth_boxes["boxes"][i][2]
        ground_truth_lower_right_y = ground_truth_boxes["boxes"][i][3]
        ground_truth_class_index = ground_truth_boxes["labels"][i] - 1

        # Draw ground truth box on input
        input = cv2.rectangle(input, (ground_truth_upper_left_x, ground_truth_upper_left_y),
            (ground_truth_lower_right_x, ground_truth_lower_right_y),
            header.ground_truth_color, header.ground_truth_thickness)

        # Read ground truth box class name and compose ground truth box text
        ground_truth_text = classes[ground_truth_class_index]

        # Obtain ground truth box text dimensions
        ground_truth_text_size = cv2.getTextSize(ground_truth_text,
            cv2.FONT_HERSHEY_SIMPLEX, header.ground_truth_text_size, header.ground_truth_thickness)[0]
        ground_truth_text_width = int(ground_truth_text_size[0]) + header.ground_truth_text_margin_x
        ground_truth_text_height = int(ground_truth_text_size[1]) + header.ground_truth_text_margin_y

        # Draw ground truth box text background on input
        input = cv2.rectangle(input, (ground_truth_upper_left_x, ground_truth_upper_left_y - ground_truth_text_height),
            (ground_truth_upper_left_x + ground_truth_text_width, ground_truth_upper_left_y),
            header.ground_truth_color, -1)

        # Draw ground truth box text on input
        input = cv2.putText(input, ground_truth_text,
            (ground_truth_upper_left_x + int(header.ground_truth_text_margin_x / 2),
            ground_truth_upper_left_y - int(header.ground_truth_text_margin_y / 1.4)), cv2.FONT_HERSHEY_SIMPLEX,
            header.ground_truth_text_size, header.ground_truth_text_color, header.ground_truth_thickness, cv2.LINE_AA)

    # Process bounding boxes
    for bounding_box in bounding_boxes:
        # Read bounding box parameters
        bounding_box_upper_left_x = bounding_box[0]
        bounding_box_upper_left_y = bounding_box[1]
        bounding_box_lower_right_x = bounding_box[2]
        bounding_box_lower_right_y = bounding_box[3]
        bounding_box_class_score = bounding_box[5] * 100
        bounding_box_class_index = bounding_box[6]

        # Draw bounding box on input
        input = cv2.rectangle(input, (bounding_box_upper_left_x, bounding_box_upper_left_y),
            (bounding_box_lower_right_x, bounding_box_lower_right_y),
            header.bounding_box_color, header.bounding_box_thickness)

        # Read bounding box class name and compose bounding box text
        bounding_box_class_name = classes[bounding_box_class_index]
        bounding_box_text = bounding_box_class_name + ' ' + "{:.2f}".format(bounding_box_class_score) + '%'

        # Obtain bounding box text dimensions
        bounding_box_text_size = cv2.getTextSize(bounding_box_text,
            cv2.FONT_HERSHEY_SIMPLEX, header.bounding_box_text_size, header.bounding_box_thickness)[0]
        bounding_box_text_width = int(bounding_box_text_size[0]) + header.bounding_box_text_margin_x
        bounding_box_text_height = int(bounding_box_text_size[1]) + header.bounding_box_text_margin_y

        # Draw bounding box text background on input
        input = cv2.rectangle(input, (bounding_box_upper_left_x, bounding_box_upper_left_y - bounding_box_text_height),
            (bounding_box_upper_left_x + bounding_box_text_width, bounding_box_upper_left_y),
            header.bounding_box_color, -1)

        # Draw bounding box text on input
        input = cv2.putText(input, bounding_box_text,
            (bounding_box_upper_left_x + int(header.bounding_box_text_margin_x / 2),
            bounding_box_upper_left_y - int(header.bounding_box_text_margin_y / 1.4)), cv2.FONT_HERSHEY_SIMPLEX,
            header.bounding_box_text_size, header.bounding_box_text_color, header.bounding_box_thickness, cv2.LINE_AA)

    # Return input with boxes drawn
    return input

def calculate_output_size(input_size, kernel_size, dilation, stride, padding):
    # input_size = (in_height, in_width)
    # https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html

    # Calculate output size
    output_size_height = math.floor((input_size[0] + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1)
    output_size_width = math.floor((input_size[1] + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1)

    # Return output size
    return (output_size_height, output_size_width)

def calculate_unfold_size(input_size, kernel_size, dilation, padding, stride):
    # input_size = (batch_size, in_channels, in_height, in_width)
    # https://pytorch.org/docs/stable/generated/torch.nn.Unfold.html

    # Caculate unfold size
    # batch_size = N
    # block_size = C x kernel_size x kernel_size
    # num_blocks = L
    batch_size = int(input_size[0])
    in_channels = int(input_size[1])
    in_height = int(input_size[2])
    in_width = int(input_size[3])
    block_size = int(in_channels * kernel_size[0])
    num_blocks_shape = calculate_output_size((in_height, in_width), kernel_size, dilation, padding, stride)
    num_blocks = num_blocks_shape[0] * num_blocks_shape[1]

    # Return unfold size
    return (batch_size, block_size, num_blocks)

def format_ground_truth_bboxes(gt):
    gt_formatted_boxes = []

    for box in gt:
        gt_box = []

        box_coords = box["bbox"]

        gt_box.append(box_coords[0]) # xmin
        gt_box.append(box_coords[1]) # ymin
        gt_box.append(box_coords[0] + box_coords[2]) # xmax
        gt_box.append(box_coords[1] + box_coords[3]) # ymax
        gt_box.append(box["category_id"]) # class_id
        gt_box.append(0) # difficult

        if box["iscrowd"] == 0:
            gt_box.append(0) # crowd
        else:
            gt_box.append(0) # crowd

        gt_formatted_boxes.append(gt_box)

    return gt_formatted_boxes

def format_gt_for_mAP(file_ground_truth):
    if os.path.isfile(header.file_path_name_ground_truth_formatted):
        return

    logger.log_info("Creating formatted ground truth file...")

    output = dict()

    for key in file_ground_truth.imgs.keys():
        image_id = file_ground_truth.imgs[key]["id"]
        gt = []

        for key in file_ground_truth.anns.keys():
            if file_ground_truth.anns[key]["image_id"] == image_id:
                    gt.append(file_ground_truth.anns[key])

        formatted_ground_truth_bboxes =  format_ground_truth_bboxes(gt)
        output[image_id] = formatted_ground_truth_bboxes

    with open(header.file_path_name_ground_truth_formatted, 'w+') as f:
        json.dump(output, f)
